@@ -32,9 +32,13 @@
 #undef EXPANDED_LOGDIR
 #undef EXPANDED_SYSCONFDIR
 #undef EXPANDED_SESSDIR
+#undef X_PATH
 #undef X_SERVER
 #undef X_SERVER_PATH
+#undef X_XNEST_PATH
+#undef X_XNEST_CONFIG_OPTIONS
 #undef X_CONF_PATH
+#undef X_CONFIG_OPTIONS
 #undef GETTEXT_PACKAGE
 #undef HAVE_LOGINCAP
 #undef HAVE_SYS_SOCKIO_H
