@@ -113,6 +113,8 @@ enum {
 #define GDM_INTERRUPT_CONFIGURE   'C'
 #define GDM_INTERRUPT_SUSPEND     'S'
 #define GDM_INTERRUPT_SELECT_USER 'U'
+#define GDM_INTERRUPT_LOGIN_SOUND 'L'
+#define GDM_INTERRUPT_THEME       'H'
 
 /* The dreaded miscellaneous category */
 #define FIELD_SIZE 256
@@ -145,7 +147,7 @@ enum {
 #define GDM_KEY_REBOOT "daemon/RebootCommand=/usr/bin/reboot;/sbin/reboot;/sbin/shutdown -r now;/usr/sbin/shutdown -r now"
 #define GDM_KEY_ROOTPATH "daemon/RootPath=/sbin:/usr/sbin:/bin:/usr/bin:" X_CONF_PATH ":" EXPANDED_BINDIR
 #define GDM_KEY_SERVAUTH "daemon/ServAuthDir=" EXPANDED_AUTHDIR
-#define GDM_KEY_SESSDIR "daemon/SessionDesktopDir=/etc/X11/sessions/:" EXPANDED_SYSCONFDIR "/dm/Sessions/:" EXPANDED_DATADIR "/xsessions/"
+#define GDM_KEY_SESSDIR "daemon/SessionDesktopDir=/etc/X11/sessions/:" EXPANDED_SYSCONFDIR "/dm/Sessions/:" EXPANDED_DATADIR "/gdm/BuiltInSessions/:" EXPANDED_DATADIR "/xsessions/"
 #define GDM_KEY_BASEXSESSION "daemon/BaseXsession=" EXPANDED_SYSCONFDIR "/gdm/Xsession"
 #define GDM_KEY_DEFAULTSESSION "daemon/DefaultSession=gnome.desktop"
 #define GDM_KEY_SUSPEND "daemon/SuspendCommand="
@@ -157,6 +159,8 @@ enum {
 
 #define GDM_KEY_DOUBLELOGINWARNING "daemon/DoubleLoginWarning=true"
 
+#define GDM_KEY_DISPLAY_LAST_LOGIN "daemon/DisplayLastLogin=false"
+
 /* This defaults to true for backward compatibility,
  * it will not actually do timed login since the TimedLogin defaults
  * to nothing */
@@ -164,6 +168,8 @@ enum {
 #define GDM_KEY_TIMED_LOGIN "daemon/TimedLogin="
 #define GDM_KEY_TIMED_LOGIN_DELAY "daemon/TimedLoginDelay=30"
 
+#define GDM_KEY_FLEXI_REAP_DELAY_MINUTES "daemon/FlexiReapDelayMinutes=5"
+
 #define GDM_KEY_STANDARD_XSERVER "daemon/StandardXServer=" X_SERVER
 #define GDM_KEY_FLEXIBLE_XSERVERS "daemon/FlexibleXServers=5"
 #define GDM_KEY_XNEST "daemon/Xnest=" X_SERVER_PATH "/Xnest -name Xnest"
@@ -172,6 +178,8 @@ enum {
 #define GDM_KEY_FIRSTVT "daemon/FirstVT=7"
 #define GDM_KEY_VTALLOCATION "daemon/VTAllocation=true"
 
+#define GDM_KEY_CONSOLE_CANNOT_HANDLE "daemon/ConsoleCannotHandle=am,ar,az,bn,el,fa,gu,hi,ja,ko,ml,mr,pa,ta,zh"
+
 /* Per server definitions */
 #define GDM_KEY_SERVER_NAME "name=Standard server"
 #define GDM_KEY_SERVER_COMMAND "command=" X_SERVER
@@ -189,9 +197,12 @@ enum {
 #define GDM_KEY_ALLOWREMOTEAUTOLOGIN "security/AllowRemoteAutoLogin=false"
 #define GDM_KEY_MAXFILE "security/UserMaxFile=65536"
 #define GDM_KEY_RELAXPERM "security/RelaxPermissions=0"
-#define GDM_KEY_RETRYDELAY "security/RetryDelay=3"
+#define GDM_KEY_CHECKDIROWNER "security/CheckDirOwner=true"
+#define GDM_KEY_RETRYDELAY "security/RetryDelay=1"
 #define GDM_KEY_DISALLOWTCP "security/DisallowTCP=true"
 
+#define GDM_KEY_NEVERPLACECOOKIESONNFS "security/NeverPlaceCookiesOnNFS=true"
+
 #define GDM_KEY_XDMCP "xdmcp/Enable=false"
 #define GDM_KEY_MAXPEND "xdmcp/MaxPending=4"
 #define GDM_KEY_MAXSESS "xdmcp/MaxSessions=16"
@@ -204,10 +215,14 @@ enum {
 #define GDM_KEY_PINGINTERVAL "xdmcp/PingIntervalSeconds=15"
 #define GDM_KEY_WILLING "xdmcp/Willing=" EXPANDED_SYSCONFDIR "/gdm/Xwilling"
 
-#define GDM_KEY_GTKRC "gui/GtkRC=" EXPANDED_DATADIR "/themes/Default/gtk/gtkrc"
+#define GDM_KEY_GTK_THEME "gui/GtkTheme=Default"
+#define GDM_KEY_GTKRC "gui/GtkRC="
 #define GDM_KEY_ICONWIDTH "gui/MaxIconWidth=128"
 #define GDM_KEY_ICONHEIGHT "gui/MaxIconHeight=128"
 
+#define GDM_KEY_ALLOW_GTK_THEME_CHANGE "gui/AllowGtkThemeChange=true"
+#define GDM_KEY_GTK_THEMES_TO_ALLOW "gui/GtkThemesToAllow=all"
+
 #define GDM_KEY_BROWSER "greeter/Browser=false"
 #define GDM_KEY_EXCLUDE "greeter/Exclude=bin,daemon,adm,lp,sync,shutdown,halt,mail,news,uucp,operator,nobody,gdm,postgres,pvm,rpm,nfsnobody,pcap"
 #define GDM_KEY_MINIMALUID "greeter/MinimalUID=100"
@@ -227,7 +242,7 @@ enum {
 #define GDM_KEY_BACKGROUNDPROG "greeter/BackgroundProgram="
 #define GDM_KEY_RUNBACKGROUNDPROGALWAYS "greeter/RunBackgroundProgramAlways=false"
 #define GDM_KEY_BACKGROUNDIMAGE "greeter/BackgroundImage="
-#define GDM_KEY_BACKGROUNDCOLOR "greeter/BackgroundColor=#363047"
+#define GDM_KEY_BACKGROUNDCOLOR "greeter/BackgroundColor=#76848F"
 #define GDM_KEY_BACKGROUNDTYPE "greeter/BackgroundType=2"
 #define GDM_KEY_BACKGROUNDSCALETOFIT "greeter/BackgroundScaleToFit=true"
 #define GDM_KEY_BACKGROUNDREMOTEONLYCOLOR "greeter/BackgroundRemoteOnlyColor=true"
@@ -240,10 +255,21 @@ enum {
 #define GDM_KEY_GRAPHICAL_THEME "greeter/GraphicalTheme=circles"
 #define GDM_KEY_GRAPHICAL_THEME_DIR "greeter/GraphicalThemeDir=" EXPANDED_DATADIR "/gdm/themes/"
 
+#define GDM_KEY_INFO_MSG_FILE "greeter/InfoMsgFile="
+#define GDM_KEY_INFO_MSG_FONT "greeter/InfoMsgFont="
+
+#define GDM_KEY_SOUND_ON_LOGIN "greeter/SoundOnLogin=true"
+#define GDM_KEY_SOUND_ON_LOGIN_FILE "greeter/SoundOnLoginFile="
+#define GDM_KEY_SOUND_PROGRAM "daemon/SoundProgram=/usr/bin/play"
+
 #define GDM_KEY_SCAN "chooser/ScanTime=4"
 #define GDM_KEY_HOST "chooser/DefaultHostImg=" EXPANDED_PIXMAPDIR "/nohost.png"
 #define GDM_KEY_HOSTDIR "chooser/HostImageDir=" EXPANDED_DATADIR "/hosts/"
 #define GDM_KEY_HOSTS "chooser/Hosts="
+#ifdef ENABLE_IPV6
+#define GDM_KEY_MULTICAST "chooser/Multicast=true"
+#define GDM_KEY_MULTICAST_ADDR "chooser/MulticastAddr=ff02::1"
+#endif
 #define GDM_KEY_BROADCAST "chooser/Broadcast=true"
 #define GDM_KEY_ALLOWADD "chooser/AllowAdd=true"
 
@@ -270,7 +296,18 @@ enum {
 typedef struct _GdmConnection GdmConnection;
 #endif  /* TYPEDEF_GDM_CONNECTION */
 
+typedef enum {
+	GDM_LOGOUT_ACTION_NONE = 0,
+	GDM_LOGOUT_ACTION_HALT,
+	GDM_LOGOUT_ACTION_REBOOT,
+	GDM_LOGOUT_ACTION_SUSPEND,
+	GDM_LOGOUT_ACTION_LAST
+} GdmLogoutAction;
+
+#ifndef TYPEDEF_GDM_DISPLAY
+#define TYPEDEF_GDM_DISPLAY
 typedef struct _GdmDisplay GdmDisplay;
+#endif /* TYPEDEF_GDM_DISPLAY */
 
 /* Use this to get the right authfile name */
 #define GDM_AUTHFILE(display) \
@@ -286,6 +323,7 @@ struct _GdmDisplay {
     GSList *local_auths; 
     gchar *userauth;
     gboolean authfb;
+    time_t last_auth_touch;
     gchar *command;
     gboolean failsafe_xserver;
     gboolean use_chooser; /* run chooser instead of greeter */
@@ -297,7 +335,13 @@ struct _GdmDisplay {
     gchar *name;
     gchar *hostname;
     struct in_addr addr;
+#ifdef ENABLE_IPV6
+    struct in6_addr addr6;
+    struct sockaddr_storage *addrs; /* array of addresses */
+#else
     struct in_addr *addrs; /* array of addresses */
+#endif
+    int addrtype;        /* Specifying the variable used, addr or addr6  */
     int addr_count; /* number of addresses in array */
     /* Note that the above may in fact be empty even though
        addr is set, these are just extra addresses
@@ -373,6 +417,14 @@ struct _GdmDisplay {
 						the ~/.xsession-errors file */
     char *xsession_errors_filename; /* if NULL then there is no .xsession-errors
 				       file */
+
+    int chooser_output_fd; /* from the chooser */
+    char *chooser_last_line;
+
+    /* Only set in the main daemon as that's the only place that cares */
+    GdmLogoutAction logout_action;
+    
+    char *theme_name;
 };
 
 typedef struct _GdmXServer GdmXServer;
@@ -389,7 +441,12 @@ struct _GdmXServer {
 typedef struct _GdmIndirectDisplay GdmIndirectDisplay;
 struct _GdmIndirectDisplay {
 	int id;
+#ifdef ENABLE_IPV6
+       struct sockaddr_storage* dsp_sa;
+       struct in6_addr* chosen_host6;
+#else
 	struct sockaddr_in* dsp_sa;
+#endif
 	time_t acctime;
 	struct in_addr *chosen_host;
 };
@@ -398,8 +455,13 @@ struct _GdmIndirectDisplay {
 typedef struct _GdmForwardQuery GdmForwardQuery;
 struct _GdmForwardQuery {
 	time_t acctime;
+#ifdef ENABLE_IPV6
+       struct sockaddr_storage* dsp_sa;
+       struct sockaddr_storage* from_sa;
+#else
 	struct sockaddr_in* dsp_sa;
 	struct sockaddr_in* from_sa;
+#endif
 };
 #define GDM_MAX_FORWARD_QUERIES 10
 #define GDM_FORWARD_QUERY_TIMEOUT 30
@@ -470,16 +532,24 @@ void		gdm_final_cleanup	(void);
 #define GDM_SOP_SOFT_RESTART "SOFT_RESTART" /* no arguments */
 #define GDM_SOP_START_NEXT_LOCAL "START_NEXT_LOCAL" /* no arguments */
 #define GDM_SOP_HUP_ALL_GREETERS "HUP_ALL_GREETERS" /* no arguments */
+
+/* stop waiting for this and go on with your life, useful with
+   the --wait-for-go command line option */
+#define GDM_SOP_GO "GO" /* no arguments */
+
 /* sometimes we can't do a syslog so we tell the main daemon */
 #define GDM_SOP_SYSLOG "SYSLOG" /* <pid> <type> <message> */
+
 /* write out a sessreg (xdm) compatible Xservers file
  * in the ServAuthDir as <name>.Xservers */
 #define GDM_SOP_WRITE_X_SERVERS "WRITE_X_SERVERS" /* <slave pid> */
+
 /* All X servers should be restarted rather then regenerated.  Useful
  * if you have updated the X configuration.  Note that this happens
  * only when the user logs out or when we otherwise would have restarted
  * a server, nothing is done by this command. */
 #define GDM_SOP_DIRTY_SERVERS "DIRTY_SERVERS"  /* no arguments */
+
 /* restart all servers that people aren't logged in on.  Maybe you may not
  * want to do this on every change of X server config since this may cause
  * flicker on screen and jumping around on the vt.  Perhaps useful to do
@@ -488,6 +558,10 @@ void		gdm_final_cleanup	(void);
 #define GDM_SOP_SOFT_RESTART_SERVERS "SOFT_RESTART_SERVERS"  /* no arguments */
 /* Suspend the machine if it is even allowed */
 #define GDM_SOP_SUSPEND_MACHINE "SUSPEND_MACHINE"  /* no arguments */
+#define GDM_SOP_CHOSEN_THEME "CHOSEN_THEME"  /* <slave pid> <theme name> */
+
+/* Start a new standard X flexible server */
+#define GDM_SOP_FLEXI_XSERVER "FLEXI_XSERVER" /* no arguments */
 
 /* Notification protocol */
 /* keys */
@@ -503,10 +577,15 @@ void		gdm_final_cleanup	(void);
 #define GDM_NOTIFY_TIMED_LOGIN "TimedLogin" /* <login> */
 #define GDM_NOTIFY_TIMED_LOGIN_DELAY "TimedLoginDelay" /* <seconds> */
 #define GDM_NOTIFY_DISALLOWTCP "DisallowTCP" /* <true/false as int> */
+#define GDM_NOTIFY_SOUND_ON_LOGIN_FILE "SoundOnLoginFile" /* <sound file> */
+#define GDM_NOTIFY_ADD_GTK_MODULES "AddGtkModules" /* <true/false as int> */
+#define GDM_NOTIFY_GTK_MODULES_LIST "GtkModulesList" /* <modules list> */
 
-/* commands */
+/* commands, seel GDM_SLAVE_NOTIFY_COMMAND */
 #define GDM_NOTIFY_DIRTY_SERVERS "DIRTY_SERVERS"
 #define GDM_NOTIFY_SOFT_RESTART_SERVERS "SOFT_RESTART_SERVERS"
+#define GDM_NOTIFY_GO "GO"
+#define GDM_NOTIFY_TWIDDLE_POINTER "TWIDDLE_POINTER"
 
 /* Ack for a slave message */
 /* Note that an extra response can follow an 'ack' */
@@ -526,6 +605,9 @@ void		gdm_final_cleanup	(void);
  * is the gdm version and not a "protocol" revision, so you can't check
  * against a single version but check if the version is higher then some
  * value.
+ *
+ * You can only send a few commands at a time, so if you keep getting error
+ * 200 try opening a new socket for every command you send.
  */
 /* The user protocol, using /tmp/.gdm_socket */
 
@@ -535,6 +617,9 @@ void		gdm_final_cleanup	(void);
  * Arguments:  None
  * Answers:
  *   GDM <gdm version>
+ *   ERROR <err number> <english error description>
+ *      200 = Too many messages
+ *      999 = Unknown error
  */
 #define GDM_SUP_AUTH_LOCAL "AUTH_LOCAL" /* <xauth cookie> */
 /* AUTH_LOCAL: Setup this connection as authenticated for FLEXI_SERVER
@@ -543,6 +628,11 @@ void		gdm_final_cleanup	(void);
  *             only users logged in from gdm.  They must pass the xauth
  *             MIT-MAGIC-COOKIE-1 that they were passed before the
  *             connection is authenticated.
+ *             Note that since 2.6.0.6 you can also use a global
+ *             <ServAuthDir>/.cookie, which works for all authentication
+ *             except for SET_LOGOUT_ACTION and QUERY_LOGOUT_ACTION
+ *             and SET_SAFE_LOGOUT_ACTION which require a logged in
+ *             display
  * Supported since: 2.2.4.0
  * Arguments:  <xauth cookie>
  *   <xauth cookie> is in hex form with no 0x prefix
@@ -551,6 +641,7 @@ void		gdm_final_cleanup	(void);
  *   ERROR <err number> <english error description>
  *      0 = Not implemented
  *      100 = Not authenticated
+ *      200 = Too many messages
  *      999 = Unknown error
  */
 #define GDM_SUP_FLEXI_XSERVER "FLEXI_XSERVER" /* <xserver type> */
@@ -569,6 +660,7 @@ void		gdm_final_cleanup	(void);
  *      4 = X too busy
  *      6 = No server binary
  *      100 = Not authenticated
+ *      200 = Too many messages
  *      999 = Unknown error
  */
 #define GDM_SUP_FLEXI_XNEST  "FLEXI_XNEST" /* <display> <uid> <xauth cookie> <xauth file> */
@@ -597,6 +689,7 @@ void		gdm_final_cleanup	(void);
  *      5 = Xnest can't connect
  *      6 = No server binary
  *      100 = Not authenticated
+ *      200 = Too many messages
  *      999 = Unknown error
  */
 #define GDM_SUP_CONSOLE_SERVERS  "CONSOLE_SERVERS" /* None */
@@ -614,6 +707,11 @@ void		gdm_final_cleanup	(void);
  *   for example).  If the display is an xnest display and is a console one
  *   (that is, it is an xnest inside another console display) it is listed
  *   and instead of vt, it lists the parent display in standard form.
+ *
+ *   ERROR <err number> <english error description>
+ *      1 = Not implemented
+ *      200 = Too many messages
+ *      999 = Unknown error
  */
 #define GDM_SUP_ALL_SERVERS  "ALL_SERVERS" /* None */
 /* ALL_SERVERS: List all servers, including console, remote, xnest.  This
@@ -628,6 +726,11 @@ void		gdm_final_cleanup	(void);
  *   <server> is <display>,<logged in user>
  *
  *   <logged in user> can be empty in case no one logged in yet
+ *
+ *   ERROR <err number> <english error description>
+ *      0 = Not implemented
+ *      200 = Too many messages
+ *      999 = Unknown error
  */
 #define GDM_SUP_UPDATE_CONFIG "UPDATE_CONFIG" /* <key> */
 /* UPDATE_CONFIG: Tell the daemon to update config of some key.  Any user
@@ -653,12 +756,15 @@ void		gdm_final_cleanup	(void);
  *			xdmcp/MaxPendingIndirect
  *			xdmcp/MaxWaitIndirect
  *			xdmcp/PingIntervalSeconds (only affects new connections)
- *   		 xdmcp/TimedLogin (2.3.90.3)
- *   		 xdmcp/TimedLoginEnable (2.3.90.3)
- *   		 xdmcp/TimedLoginDelay (2.3.90.3)
+ *   		 daemon/TimedLogin (2.3.90.3)
+ *   		 daemon/TimedLoginEnable (2.3.90.3)
+ *   		 daemon/TimedLoginDelay (2.3.90.3)
  *   		 greeter/SystemMenu (2.3.90.3)
  *   		 greeter/ConfigAvailable (2.3.90.3)
  *   		 greeter/ChooserButton (2.4.2.0)
+ *               greeter/SoundOnLoginFile (2.5.90.0)
+ *               daemon/AddGtkModules (2.5.90.0)
+ *               daemon/GtkModulesList (2.5.90.0)
  * Supported since: 2.3.90.2
  * Arguments:  <key>
  *   <key> is just the base part of the key such as "security/AllowRemoteRoot"
@@ -667,6 +773,7 @@ void		gdm_final_cleanup	(void);
  *   ERROR <err number> <english error description>
  *      0 = Not implemented
  *      50 = Unsupported key
+ *      200 = Too many messages
  *      999 = Unknown error
  */
 #define GDM_SUP_GREETERPIDS  "GREETERPIDS" /* None */
@@ -676,8 +783,115 @@ void		gdm_final_cleanup	(void);
  * Arguments:  None
  * Answers:
  *   OK <pid>;<pid>;...
+ *   ERROR <err number> <english error description>
+ *      0 = Not implemented
+ *      200 = Too many messages
+ *      999 = Unknown error
  */
-#define GDM_SUP_CLOSE        "CLOSE" /* no arguments */
+#define GDM_SUP_QUERY_LOGOUT_ACTION "QUERY_LOGOUT_ACTION" /* None */
+/* QUERY_LOGOUT_ACTION: Query which logout actions are possible
+ * Only supported on connections that passed AUTH_LOCAL.
+ * Supported since: 2.5.90.0
+ * Answers:
+ *   OK <action>;<action>;...
+ *      Where action is one of HALT, REBOOT or SUSPEND.  An empty list
+ *      can also be returned if no action is possible.  A '!' is appended
+ *      to an action if it was already set with SET_LOGOUT_ACTION or
+ *      SET_SAFE_LOGOUT_ACTION.  Note that SET_LOGOUT_ACTION has precedence
+ *      over SET_SAFE_LOGOUT_ACTION.
+ *   ERROR <err number> <english error description>
+ *      0 = Not implemented
+ *      100 = Not authenticanted
+ *      200 = Too many messages
+ *      999 = Unknown error
+ */
+#define GDM_SUP_SET_LOGOUT_ACTION "SET_LOGOUT_ACTION" /* <action> */
+/* SET_LOGOUT_ACTION:  Tell the daemon to halt/reboot/suspend after slave
+ * process exits. 
+ * Only supported on connections that passed AUTH_LOCAL.
+ * Supported since: 2.5.90.0
+ * Arguments:  <action>
+ *   NONE           Set exit action to 'none'
+ *   HALT           Set exit action to 'halt'
+ *   REBOOT         Set exit action to 'reboot'
+ *   SUSPEND        Set exit action to 'suspend'
+ *
+ * Answers:
+ *   OK
+ *   ERROR <err number> <english error description>
+ *      0 = Not implemented
+ *      7 = Unknown logout action, or not available
+ *      100 = Not authenticanted
+ *      200 = Too many messages
+ *      999 = Unknown error
+ */
+#define GDM_SUP_SET_SAFE_LOGOUT_ACTION "SET_SAFE_LOGOUT_ACTION" /* <action> */
+/* SET_SAFE_LOGOUT_ACTION:  Tell the daemon to halt/reboot/suspend after
+ * everybody logs out.  If only one person logs out, then this is obviously
+ * the same as the SET_LOGOUT_ACTION.  Note that SET_LOGOUT_ACTION has
+ * precendence over SET_SAFE_LOGOUT_ACTION if it is set to something other
+ * then NONE.  If no one is logged in, then the action takes effect
+ * immedeately.
+ * Only supported on connections that passed AUTH_LOCAL.
+ * Supported since: 2.5.90.0
+ * Arguments:  <action>
+ *   NONE           Set exit action to 'none'
+ *   HALT           Set exit action to 'halt'
+ *   REBOOT         Set exit action to 'reboot'
+ *   SUSPEND        Set exit action to 'suspend'
+ *
+ * Answers:
+ *   OK
+ *   ERROR <err number> <english error description>
+ *      0 = Not implemented
+ *      7 = Unknown logout action, or not available
+ *      100 = Not authenticanted
+ *      200 = Too many messages
+ *      999 = Unknown error
+ */
+#define GDM_SUP_LOGOUT_ACTION_NONE	"NONE"
+#define GDM_SUP_LOGOUT_ACTION_HALT	"HALT"
+#define GDM_SUP_LOGOUT_ACTION_REBOOT	"REBOOT"
+#define GDM_SUP_LOGOUT_ACTION_SUSPEND	"SUSPEND"
+/*
+ */
+#define GDM_SUP_QUERY_VT "QUERY_VT" /* None */
+/* QUERY_VT:  Ask the daemon about which VT we are currently on.
+ * This is useful for logins which don't own /dev/console but are
+ * still console logins.  Only supported on Linux currently, other places
+ * will just get ERROR 8.  This is also the way to query if VT
+ * support is available in the daemon in the first place.
+ * Only supported on connections that passed AUTH_LOCAL.
+ * Supported since: 2.5.90.0
+ * Arguments:  None
+ * Answers:
+ *   OK <vt number>
+ *   ERROR <err number> <english error description>
+ *      0 = Not implemented
+ *      8 = Virtual terminals not supported
+ *      100 = Not authenticanted
+ *      200 = Too many messages
+ *      999 = Unknown error
+ */
+#define GDM_SUP_SET_VT "SET_VT" /* <vt> */
+/* SET_VT:  Change to the specified virtual terminal.
+ * This is useful for logins which don't own /dev/console but are
+ * still console logins.  Only supported on Linux currently, other places
+ * will just get ERROR 8.
+ * Only supported on connections that passed AUTH_LOCAL.
+ * Supported since: 2.5.90.0
+ * Arguments:  None
+ * Answers:
+ *   OK
+ *   ERROR <err number> <english error description>
+ *      0 = Not implemented
+ *      8 = Virtual terminals not supported
+ *      9 = Invalid virtual terminal number
+ *      100 = Not authenticanted
+ *      200 = Too many messages
+ *      999 = Unknown error
+ */
+#define GDM_SUP_CLOSE        "CLOSE" /* None */
 /* CLOSE Answers: None
  * Supported since: 2.2.4.0
  */
@@ -688,29 +902,36 @@ enum {
 					  * from a local display we started */
 };
 
-#define IGNORE_EINTR(expr) \
-	do {		\
-		errno = 0;	\
-		expr;		\
-	} while G_UNLIKELY (errno == EINTR);
 
 #define NEVER_FAILS_seteuid(uid) \
-	{ int r = seteuid (uid); \
+	{ int r = 0; \
+	  if (geteuid () != uid) \
+	    r = seteuid (uid); \
 	  if G_UNLIKELY (r != 0) \
-        gdm_fail ("GDM file %s: line %d (%s): Cannot run seteuid to %d", \
+        gdm_fail ("GDM file %s: line %d (%s): Cannot run seteuid to %d: %s", \
 		  __FILE__,						\
 		  __LINE__,						\
 		  __PRETTY_FUNCTION__,					\
-                  (int)uid);			}
+                  (int)uid,						\
+		  strerror (errno));			}
 #define NEVER_FAILS_setegid(gid) \
-	{ int r = setegid (gid); \
+	{ int r = 0; \
+	  if (getegid () != gid) \
+	    r = setegid (gid); \
 	  if G_UNLIKELY (r != 0) \
-        gdm_fail ("GDM file %s: line %d (%s): Cannot run setegid to %d", \
+        gdm_fail ("GDM file %s: line %d (%s): Cannot run setegid to %d: %s", \
 		  __FILE__,						\
 		  __LINE__,						\
 		  __PRETTY_FUNCTION__,					\
-                  (int)gid);			}
-		
+                  (int)gid,						\
+		  strerror (errno));			}
+
+/* first goes to euid-root and then sets the egid and euid, to make sure
+ * this succeeds */
+#define NEVER_FAILS_root_set_euid_egid(uid,gid) \
+	{ NEVER_FAILS_seteuid (0); \
+	  NEVER_FAILS_setegid (gid); \
+	  if (uid != 0) { NEVER_FAILS_seteuid (uid); } }
 
 #endif /* GDM_H */
 
