@@ -20,11 +20,14 @@
 
 #include <ctype.h>
 #include <string.h>
+#include <syslog.h>
+#include <math.h>
 
 #include <glib.h>
 #include <gmodule.h>
 #include <gdk/gdk.h>
 #include <gtk/gtk.h>
+#include <vicious.h>
 #include <gnome.h>
 
 #include <config.h>
@@ -85,7 +88,7 @@ static guint max_crossings = 0;
 
 static void create_event_watcher ();
 static void load_bindings(gchar *path);
-static gchar * screen_exec_display_string (GdkScreen *screen);
+static gchar * screen_exec_display_string (GdkScreen *screen, const char *old);
 static gchar ** get_exec_environment (GdkScreen *screen);
 static Binding * parse_line(gchar *buf);
 static gboolean binding_already_used (Binding *binding);
@@ -93,9 +96,8 @@ BindingType get_binding_type(char c);
 BindingDirection get_binding_direction(char c);
 
 static gchar *
-screen_exec_display_string (GdkScreen *screen)
+screen_exec_display_string (GdkScreen *screen, const char *old)
 {
-#ifdef HAVE_GTK_MULTIHEAD
   GString    *str;
   const gchar *old_display;
   gchar       *retval;
@@ -119,9 +121,6 @@ screen_exec_display_string (GdkScreen *screen)
   g_string_free (str, FALSE);
 
   return retval;
-#else
-  return g_strdup ("DISPLAY=:0.0");
-#endif
 }
 
 /**
@@ -147,17 +146,17 @@ get_exec_environment (GdkScreen *screen)
   g_assert (GDK_IS_SCREEN (screen));
 
   for (i = 0; environ [i]; i++)
-    if (!strncmp (environ [i], "DISPLAY", 7))
+    if (strncmp (environ [i], "DISPLAY", 7) == 0)
       display_index = i;
 
   if (display_index == -1)
     display_index = i++;
 
-  retval = g_new (char *, i + 1);
+  retval = g_new0 (char *, i + 1);
 
   for (i = 0; environ [i]; i++)
     if (i == display_index)
-      retval [i] = screen_exec_display_string (screen);
+      retval [i] = screen_exec_display_string (screen, environ[i]);
   else
       retval [i] = g_strdup (environ [i]);
 
@@ -200,45 +199,50 @@ get_binding_direction(char c)
   return rc;
 }
 
+static void
+free_binding (Binding *binding)
+{
+	if (binding == NULL)
+		return;
+	g_slist_foreach (binding->actions, (GFunc)g_free, NULL);
+	g_slist_free (binding->actions);
+	g_free (binding->binding_str);
+	g_free (binding->input.gesture);
+	g_free (binding);
+}
+
 static Binding *
 parse_line(gchar *buf)
 {
-  gchar *c;
   gchar *keystring, *keyservice;
   Binding *tmp_binding = NULL;
   static GdkDisplay *display = NULL;
+
+  lineno++;
   
   if (!display)
     {
       if ((display = gdk_display_get_default ()) == NULL)
         return NULL;
     }
-  lineno++;
 
   if ((*buf == '#') || (iseol (*buf)) || (buf == NULL))
     return NULL;
 
-  tmp_binding = g_new0 (Binding, 1);
-  keystring = c = buf;
-
   /*
    * Find the binding name
    */
-  while(!(isspace (*c)))
+  keystring = strtok (buf, " \t\n\r\f");
+  if (keystring == NULL)
     {
-      if (iseol (*c))
-        {
-          /* TODO - Error messages */
-          return NULL;
-        }
-      c++;
+      /* TODO - Add an error message */
+      return NULL;
     }
 
-  *c++ = '\0';
-  tmp_binding->binding_str = (gchar *)g_malloc (strlen (keystring) + 1);
-  strncpy (tmp_binding->binding_str, keystring, strlen (keystring) + 1);
+  tmp_binding = g_new0 (Binding, 1);
+  tmp_binding->binding_str = g_strdup (keystring);
 
-  if (strcmp (tmp_binding->binding_str, "<Add>"))
+  if (strcmp (tmp_binding->binding_str, "<Add>") != 0)
     {
       BindingType bt;
       BindingDirection bd;
@@ -256,7 +260,7 @@ parse_line(gchar *buf)
 
           if (bt == BINDING_DWELL_BORDER_ERROR)
             {
-              printf("Invalid value in binding %s\n", tmp_binding->binding_str);
+              g_warning ("Invalid value in binding %s\n", tmp_binding->binding_str);
               continue;
             }
 
@@ -269,74 +273,40 @@ parse_line(gchar *buf)
 
       /* [TODO] Need to clean up here. */
      
-      /*
-       * Skip over white space
-       */
-      do
+      tmp_string = strtok (NULL, " \t\n\r\f");
+      if (tmp_string == NULL)
         {
-          if (iseol (*c))
-            {
-              /* Add an error message */
-              return NULL;
-            }
+          /* TODO - Add an error message */
+          free_binding (tmp_binding);
+          return NULL;
         }
-      while (isspace (*c) && (c++));
-
-      tmp_string = c;
-
-      while(!(isspace (*c)))
-        {
-          if (iseol (*c))
-            {
-              /* TODO - Error messages */
-              return NULL;
-            }
-          c++;
-        }
-  
-      *c++ = '\0';
 
       bd = get_binding_direction (tmp_string[0]);
 
       if (bd == BINDING_DWELL_DIRECTION_ERROR)
-        printf("Invalid value in binding %s\n", tmp_binding->binding_str);
+        g_warning ("Invalid value in binding %s\n", tmp_binding->binding_str);
       else
         tmp_binding->input.start_direction = bd;
 
-      /*
-       * Skip over white space
-       */
-      do
-        {
-          if (iseol (*c))
-            {
-              /* Add an error message */
-              return NULL;
-            }
-        }
-      while (isspace (*c) && (c++));
-
-      tmp_string = c;
-
      /*
       * Find the timeout duration (in ms). Timeout value is the 
       * time within which consecutive keypress actions must be performed
       * by the user before the sequence is discarded.
       */
-      while (!(isspace (*c)))
+
+      tmp_string = strtok (NULL, " \t\n\r\f");
+      if (tmp_string == NULL)
         {
-          if (!isdigit (*c))
-            {
-              /* Add an error message */
-              return NULL;
-            }
-          c++;
+          /* TODO - Add an error message */
+          free_binding (tmp_binding);
+          return NULL;
         }
 
-      *c++ = '\0';
-      if ((timeout=atoi (tmp_string)) <= 0)
+      timeout = atoi (tmp_string);
+      if (timeout <= 0)
         {
-          /* Add an error message */;
+          /* TODO - Add an error message */;
+	  free_binding (tmp_binding);
           return NULL;
         }
       tmp_binding->timeout = timeout;
@@ -345,19 +315,17 @@ parse_line(gchar *buf)
   /*
    * Find servcice. Permit blank space so arguments can be supplied.
    */
-  do
+  keyservice = strtok (NULL, "\n\r\f");
+  if (keyservice == NULL)
     {
-      if (iseol (*c))
-        {
-          /* Add an error message */
-          return NULL;
-        }
+      /* TODO - Add an error message */
+      free_binding (tmp_binding);
+      return NULL;
     }
-  while (isspace (*c) && (c++));
+  /* skip over initial whitespace */
+  while (*keyservice && isspace (*keyservice))
+    keyservice++;
 
-  keyservice = c;
-  for (; !iseol (*c); c++);
-  *c = '\0';
   tmp_binding->actions = g_slist_append (tmp_binding->actions, g_strdup (keyservice));
 
   return tmp_binding;
@@ -397,14 +365,15 @@ load_bindings(gchar *path)
   Binding *tmp_binding;
   gchar buf[1024];
 
-  if ((fp = fopen (path, "r")) == NULL)
+  fp = fopen (path, "r");
+  if (fp == NULL)
     {
       /* TODO - I18n */
-      printf ("Cannot open bindings file: %s\n", path);
+      g_warning ("Cannot open bindings file: %s\n", path);
       return;
     }
 
-  while (((fgets (buf, 1024, fp)) != NULL) && ((feof (fp)) == 0))
+  while (fgets (buf, sizeof (buf), fp) != NULL)
     {
       tmp_binding = (Binding *)parse_line (buf);
 
@@ -413,7 +382,7 @@ load_bindings(gchar *path)
          /*
           * Is the key already associated with an existing binding?
           */
-          if (!strcmp (tmp_binding->binding_str, "<Add>"))
+          if (strcmp (tmp_binding->binding_str, "<Add>") == 0)
             {
              /*
               * Add another action to the last binding
@@ -431,14 +400,29 @@ load_bindings(gchar *path)
                  last_binding->actions = g_slist_append (last_binding->actions,
                    g_strdup ((gchar *)tmp_binding->actions->data));
                }
+	      free_binding (tmp_binding);
             /* Ignore duplicate bindings */
-            } else if (!binding_already_used (tmp_binding))
-              binding_list = g_slist_append (binding_list, tmp_binding);
+            }
+	  else if (!binding_already_used (tmp_binding))
+            binding_list = g_slist_append (binding_list, tmp_binding);
+	  else
+	    free_binding (tmp_binding);
         }
     }
   fclose (fp);
 }
 
+static gboolean
+change_cursor_back (gpointer data)
+{
+  GdkCursor *cursor = gdk_cursor_new (GDK_LEFT_PTR);
+  gdk_window_set_cursor (gdk_get_default_root_window (), cursor);
+  gdk_cursor_unref (cursor);
+
+  return FALSE;
+}
+
+
 static gboolean
 leave_enter_emission_hook (GSignalInvocationHint        *ihint,
                            guint                        n_param_values,
@@ -467,7 +451,7 @@ leave_enter_emission_hook (GSignalInvocationHint        *ihint,
       mid_y = rect.y + (rect.height / 2);
 
       /* avoid division by 0 */
-      if (event->x_root - mid_x == 0.0)
+      if (fabs (event->x_root - mid_x) <= 0.001)
         {
           if (event->x_root < mid_x)
             crossings[crossings_position].type = BINDING_DWELL_BORDER_LEFT;
@@ -509,7 +493,10 @@ leave_enter_emission_hook (GSignalInvocationHint        *ihint,
       for (li = binding_list; li != NULL; li = li->next)
         {
           Binding *curr_binding = (Binding *) li->data;
-          int start_position = (crossings_position - curr_binding->input.num_gestures + 1) % max_crossings;
+          int start_position = (crossings_position - curr_binding->input.num_gestures + 1 + max_crossings) % max_crossings;
+	  /* being anal here */
+	  if (start_position < 0)
+		  start_position = 0;
 
           /* check initial crossing direction */
           if (curr_binding->input.start_direction == crossings[start_position].direction)
@@ -539,7 +526,6 @@ leave_enter_emission_hook (GSignalInvocationHint        *ihint,
               /* If this is true, then gesture was recognized */
               if (i == curr_binding->input.num_gestures)
                 {
-                  GError* error = NULL;
                   gboolean retval;
                   gchar **argv = NULL;
                   gchar **envp = NULL;
@@ -547,13 +533,26 @@ leave_enter_emission_hook (GSignalInvocationHint        *ihint,
                   for (act_li = curr_binding->actions; act_li != NULL; act_li = act_li->next)
                     {
                       gchar *action = (gchar *)act_li->data;
+		      char *oldpath, *newpath;
                       g_return_val_if_fail (action != NULL, TRUE);
 
-                      if (!g_shell_parse_argv (action, NULL, &argv, &error))
-                        return TRUE;
+                      if (!g_shell_parse_argv (action, NULL, &argv, NULL))
+		        continue;
 
                       envp = get_exec_environment (gtk_window_get_screen(window));
 
+		      /* add our BINDIR to the path */
+		      oldpath = g_strdup (g_getenv ("PATH"));
+		      if (ve_string_empty (oldpath))
+		        newpath = g_strdup (EXPANDED_BINDIR);
+		      else 
+		        newpath = g_strconcat (oldpath,
+					       ":",
+					       EXPANDED_BINDIR,
+					       NULL);
+		      ve_setenv ("PATH", newpath, TRUE);
+		      g_free (newpath);
+
                       retval = g_spawn_async (NULL,
                                               argv,
                                               envp,
@@ -561,11 +560,18 @@ leave_enter_emission_hook (GSignalInvocationHint        *ihint,
                                               NULL,
                                               NULL,
                                               NULL,
-                                              &error);
+                                              NULL);
+
+		      if (ve_string_empty (oldpath))
+		        ve_unsetenv ("PATH");
+		      else
+			ve_setenv ("PATH", oldpath, TRUE);
+		      g_free (oldpath);
+
                       g_strfreev (argv);
                       g_strfreev (envp);
 
-                      if (!retval)
+                      if ( ! retval)
                         {
                           GtkWidget *dialog =
                             gtk_message_dialog_new (NULL, 0, GTK_MESSAGE_ERROR,
@@ -580,6 +586,16 @@ leave_enter_emission_hook (GSignalInvocationHint        *ihint,
                                             G_CALLBACK (gtk_widget_destroy),
                                             NULL);
                           gtk_widget_show (dialog);
+			}
+		      else
+		        {
+			  GdkCursor *cursor = gdk_cursor_new (GDK_WATCH);
+			  gdk_window_set_cursor (gdk_get_default_root_window (),
+						 cursor);
+			  gdk_cursor_unref (cursor);
+			  g_timeout_add (2000,
+					 change_cursor_back,
+					 NULL);
                         }
                     }
                 }
@@ -592,7 +608,8 @@ leave_enter_emission_hook (GSignalInvocationHint        *ihint,
   return TRUE;
 }
 
-static void create_event_watcher ()
+static void
+create_event_watcher (void)
 {
   GdkDisplay *display;
   gint i;
@@ -604,7 +621,7 @@ static void create_event_watcher ()
 
   load_bindings(CONFIGFILE);
 
-  crossings = g_new(Crossings, max_crossings);
+  crossings = g_new0(Crossings, max_crossings);
 
   for (i=0; i < max_crossings; i++)
     {
@@ -622,8 +639,6 @@ static void create_event_watcher ()
     leave_enter_emission_hook, NULL, (GDestroyNotify) NULL); 
   g_signal_add_emission_hook (leave_signal_id, 0,
     leave_enter_emission_hook, NULL, (GDestroyNotify) NULL); 
-
-  return;
 }
 
 /* The init function for this gtk module */
