@@ -2,6 +2,7 @@
  *
  * (c) 2000 Eazel, Inc.
  * (c) 2001 George Lebl
+ * (c) 2003 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -26,5 +27,13 @@ void		ve_entry_set_red (GtkWidget *w,
 
 GtkWidget *	ve_gtk_option_menu_get_item (GtkOptionMenu *option_menu,
 					     int index);
+GtkWidget *	ve_hig_dialog_new (GtkWindow      *parent,
+				   GtkDialogFlags flags,
+				   GtkMessageType type,
+				   GtkButtonsType buttons,
+				   gboolean       markup,
+				   const gchar    *header,
+				   const gchar    *messagefmt,
+				   ...) G_GNUC_PRINTF (7, 8);
 
 #endif /* VE_MISCUI_H */
